#define __KERNEL__
#define MODULE
#define EXPORT_SYMTAB
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/string.h>
#include <linux/locks.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <net/scm.h>
#include <net/ip.h>
#include <asm/uaccess.h>

static int debug = 0;

MODULE_AUTHOR("Vladimir E. Spiridenkov");
MODULE_DESCRIPTION("ftpfs");
MODULE_PARM(debug,"b");

#define FTPFS_SUPER_MAGIC 0xf14f
#define FTPFS_MAGIC 0xf14f

int sock_send(struct socket * sock, char * buf, int len) {
        mm_segment_t fs;
        struct iovec iov;
        struct msghdr msg;
        struct scm_cookie scm;
        int err;

        fs = get_fs();
        set_fs(get_ds());

        msg.msg_name = NULL;
        msg.msg_namelen = 0;
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;
        msg.msg_control = NULL;
        msg.msg_controllen = 0;
	
        iov.iov_base = (void *)buf;
        iov.iov_len = len;

        msg.msg_flags = 0;

        err = scm_send(sock, &msg, &scm);
        if (err >= 0)
	{
		err = sock->ops->sendmsg(sock, &msg, len, &scm);
		scm_destroy(&scm);
	}

        set_fs(fs);

	return err;
}

int sock_recv(struct socket * sock, char * buf, int size) {
        mm_segment_t fs;
	struct iovec iov;
	struct msghdr msg;
	struct scm_cookie scm;

        fs = get_fs();
        set_fs(get_ds());

	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	iov.iov_base = buf;
	iov.iov_len = size;
	
	memset(&scm, 0,sizeof(scm));
	size = sock->ops->recvmsg(sock, &msg, size, 0, &scm);
	if(size >= 0)
		scm_recv(sock,&msg,&scm,0);

        set_fs(fs);
        
	return size;
}

struct ftp_info_type {
        struct sockaddr_in control_address;
        char user[64];
        char pass[64];
        struct socket * control_sock;
        struct socket * port_sock;
        void * control_buf;
        size_t control_buf_size;
        size_t control_buf_off;
        size_t control_buf_len;
        void * port_buf;
        size_t port_buf_size;
        size_t port_buf_off;
        size_t port_buf_len;
};

int ftp_send_control_data(struct ftp_info_type * ftp_info, char * data) {
        int retry = 0;
        int len = strlen(data);
        int size;
rep1:
        if ((size = sock_send(ftp_info->control_sock, data, len)) < 0) {
                printk(KERN_ERR "FTPFS: *** (---) error send command\n");
                return size;
        }
        if (size != len) {
                if (retry++ < 5) {
                        printk(KERN_ERR "FTPFS: *** (---) retry %i send command\n",retry);
                        data += size;
                        len -= size;
                        goto rep1;
                } else {
                        printk(KERN_ERR "FTPFS: *** (---) fail send command\n");
                        return 0;
                }
        }
        return 1;
}

int ftp_recv_control_data(struct ftp_info_type * ftp_info, int * ind, int * last, char * data) {
        int retry;
        int size;
        char temp_char;
        int pos = 0;
        int len = 0;
        (*last) = 1;
        (*data) = '\000';
rep1:
        if (ftp_info->control_buf_len) {
                temp_char = *((char *)ftp_info->control_buf + ftp_info->control_buf_off);
                ftp_info->control_buf_off++;
                ftp_info->control_buf_len--;
                pos++;                
                if (pos == 1) {
                        if (temp_char == ' ') {
                                (*last) = 0;
                                pos = 4;
                        } else
                        (*ind) = 0;
                } else if (pos == 4) {
                        if (temp_char == '-')
                                (*last) = 0;
                }
                if (pos <= 3) {
                        if ((temp_char >= '0') && (temp_char <= '7'))
                                (*ind) = (*ind)*10 + (temp_char - '0');
                        else {
                                printk(KERN_ERR "FTPFS: *** (---) invalid control index\n");        
                                pos = 4;
                        }
                }
                if (pos > 4) {
                        if (temp_char == '\n')
                                return 0;
                        else {
                                if (temp_char >= ' ') {
                                        len++;
                                        if (len < 256) {
                                                (*(data + len - 1)) = temp_char;
                                                (*(data + len)) = '\000';
                                        }
                                        if (len == 256)
                                                printk(KERN_ERR "FTPFS: *** (---) control data too big\n");
                                }
                        }
                }
                
        } else {
                ftp_info->control_buf_off = 0;
                retry = 0;
rep2:
                if ((size = sock_recv(ftp_info->control_sock, ftp_info->control_buf, ftp_info->control_buf_size)) < 0) {
                        printk(KERN_ERR "FTPFS: *** (---) error receive control data\n");
                        return size;
                }
                if (!size) {
                        if (retry++<5) {
                                printk(KERN_ERR "FTPFS: *** (---) retry %i receive control data\n",retry);
                                goto rep2;
                        } else {
                                printk(KERN_ERR "FTPFS: *** (---) error receive control data\n");
                                return 2;
                        }
                }
                ftp_info->control_buf_len = size;
        }
        goto rep1;
}

int ftp_connect(struct ftp_info_type * ftp_info) {
        int err;
        int ind;
        int last;
        char data[256];

        if (debug)
                printk(KERN_DEBUG "FTPFS: --- (---) connect %u.%u.%u.%u:%u\n",
                       ftp_info->control_address.sin_addr.s_addr & 0xff,
                       (ftp_info->control_address.sin_addr.s_addr >> 8) & 0xff,
                       (ftp_info->control_address.sin_addr.s_addr >> 16) & 0xff,
                       ftp_info->control_address.sin_addr.s_addr >> 24,
                       ntohs(ftp_info->control_address.sin_port));

        if (ftp_info->control_sock) {
                sock_release(ftp_info->control_sock);
                ftp_info->control_sock = NULL;
        }

        if ((err = sock_create(AF_INET, SOCK_STREAM, 0, &ftp_info->control_sock)) < 0 ) {
                ftp_info->control_sock = NULL;
                printk(KERN_ERR "FTPFS: *** (---) create socket error\n");
                return err;
        }

        if ((err = ftp_info->control_sock->ops->connect(ftp_info->control_sock, (struct sockaddr *)&ftp_info->control_address, sizeof(ftp_info->control_address), 0)) < 0 ) {
                printk(KERN_ERR "FTPFS: *** (---) connect error\n");
                return err;
        }

        last = 0;
        while (!last) {
                if (err = ftp_recv_control_data(ftp_info, &ind, &last, (char *)data)) {
                        printk(KERN_ERR "FTPFS: *** (---) receive failed\n");
                        return err;
                }
                if (ind != 220) {
                        printk(KERN_ERR "FTPFS: *** (---) server not ready\n");
                        return 3;
                }
        }

        if (debug)
                printk(KERN_DEBUG "FTPFS: --- (---) login: %s\n",ftp_info->user);
        sprintf(data,"USER %s\r\n",ftp_info->user);
        if (ftp_send_control_data(ftp_info, data) < 0) {
                printk(KERN_ERR "FTPFS: *** (---) send failed\n");
                return 4;
        } else {
                last = 0;
                while (!last) {
                        if (err = ftp_recv_control_data(ftp_info, &ind, &last, (char *)data)) {
                                printk(KERN_ERR "FTPFS: *** (---) receive error\n");
                                return err;
                        }
                        if (ind != 331) {
                                printk(KERN_ERR "FTPFS: *** (---) service error\n");
                                return 5;
                        }
                }
        }

        if (debug)
                printk(KERN_DEBUG "FTPFS: --- (---) password: %s\n",ftp_info->pass);
        sprintf(data,"PASS %s\r\n",ftp_info->pass);
        if (ftp_send_control_data(ftp_info, data) < 0) {
                printk(KERN_ERR "FTPFS: *** (---) send error\n");
                return 6;
        } else {
                last = 0;
                while (!last) {
                        
                        if (err = ftp_recv_control_data(ftp_info, &ind, &last, (char *)data)) {
                                printk(KERN_ERR "FTPFS: *** (---) receive error\n");
                                return err;
                        }
                        if (ind != 230) {
                                printk(KERN_ERR "FTPFS: *** (---) service error\n");
                                return 7;
                        }
                }
        }

        return 0;
}

int ftp_disconnect(struct ftp_info_type * ftp_info) {
        int ret = 0;
        int last;
        int err;
        int ind;
        char data[256];

        if (debug)
                printk(KERN_DEBUG "FTPFS: --- (---) disconnect\n");
        if (ftp_send_control_data(ftp_info, "QUIT\r\n") < 0) {
                printk(KERN_ERR "FTPFS: *** (---) send error\n");
                ret = 2;
        } else {
                last = 0;
                while (!last) {
                        if (err = ftp_recv_control_data(ftp_info, &ind, &last, (char *)data)) {
                                printk(KERN_ERR "FTPFS: *** (---) receive error\n");
                                ret = err;
                        }
                        if ((ind != 221) && (ind != 421)) {
                                printk(KERN_ERR "FTPFS: *** (---) service error\n");
                                ret = 3;
                        }
                }
        }

        if (ftp_info->control_sock) {
                sock_release(ftp_info->control_sock);
                ftp_info->control_sock = NULL;
        }

        return ret;
}

int ftp_xchg(struct ftp_info_type * ftp_info, int * ind, int * last, char * data) {
        char command[256];
        int err;
        
        sprintf(command,"%s\r\n",data);
        if (*last) {
                if (debug)
                        printk(KERN_DEBUG "FTPFS: --> (---) %s\n",data);
                if ((err = ftp_send_control_data(ftp_info, command)) < 0)
                        return err;
                if (err = ftp_recv_control_data(ftp_info, ind, last, (char *)data))
                        return err;
                if ((*ind) == 421) {
                        if (debug)
                                printk(KERN_DEBUG "FTPFS: <-- (%i) %s\n", (*ind), data);
                        if ((err = ftp_connect(ftp_info)) != 0)
                                return err;
                        printk(KERN_DEBUG "FTPFS: --> (---) last command\n");
                        if ((err = ftp_send_control_data(ftp_info, command)) < 0)
                                return err;
                        if (err = ftp_recv_control_data(ftp_info, ind, last, (char *)data))
                                return err;
                }
        } else {
                if (err = ftp_recv_control_data(ftp_info, ind, last, data))
                        return err;
                if ((*ind) == 421) {
                        if (debug)
                                printk(KERN_DEBUG "FTPFS: <-- (%i) %s\n", (*ind), data);
                        if ((err = ftp_connect(ftp_info)) != 0)
                                return err;
                        if (err = ftp_recv_control_data(ftp_info, ind, last, (char *)data))
                                return err;
                }
        }
        if (debug)
                printk(KERN_DEBUG "FTPFS: <-- (%i) %s\n", (*ind), data);
        return 0;
}

int ftp_open(struct ftp_info_type * ftp_info, struct socket ** port_sock, char mode, unsigned rest, char * command) {
        int err;
        int ind;
        int last;
        char data[256];
        struct sockaddr_in control_address;
        struct sockaddr_in port_address;
        int address_len;
        
        port_address.sin_addr.s_addr = INADDR_ANY;
        port_address.sin_port = 0;

        if ((err = sock_create(AF_INET, SOCK_STREAM, 0, &ftp_info->port_sock)) < 0 ) {
                printk(KERN_ERR "FTPFS: *** (---) create socket error\n");
                return err;
        }

        if ((err = ftp_info->port_sock->ops->bind(ftp_info->port_sock, (struct sockaddr *)&port_address, sizeof(port_address))) < 0 ) {
                printk(KERN_ERR "FTPFS: *** (---) bind error\n");
                return err;
        }

        if ((err = ftp_info->port_sock->ops->listen(ftp_info->port_sock, 5)) < 0) {
                printk(KERN_ERR "FTPFS: *** (---) listen error\n");
                return err;
        }

        address_len = sizeof(control_address);
        if ((err = ftp_info->control_sock->ops->getname(ftp_info->control_sock, (struct sockaddr *)&control_address, &address_len, 0)) < 0) {
                printk(KERN_ERR "FTPFS: *** (---) getsockname error\n");        
                return err;
                
        }

        address_len = sizeof(port_address);
        if ((err = ftp_info->port_sock->ops->getname(ftp_info->port_sock, (struct sockaddr *)&port_address, &address_len, 0)) < 0) {
                printk(KERN_ERR "FTPFS: *** (---) getsockname error\n");        
                return err;
        }
        port_address.sin_addr.s_addr = control_address.sin_addr.s_addr;

        sprintf(data,
                "PORT %u,%u,%u,%u,%u,%u",
                port_address.sin_addr.s_addr & 0xff,
                (port_address.sin_addr.s_addr >> 8) & 0xff,
                (port_address.sin_addr.s_addr >> 16) & 0xff,
                port_address.sin_addr.s_addr >> 24,
                ntohs(port_address.sin_port) >>8 ,
                ntohs(port_address.sin_port) & 0xff);
        last = 1;
        do {
                if ((err = ftp_xchg(ftp_info, &ind, &last, (char *)data)) < 0 )
                        return err;
                if (ind != 200) {
                        printk(KERN_ERR "FTPFS: *** (---) service error\n");
                        return 8;
                }
        } while (!last);

        sprintf(data, "TYPE %c", mode);
        last = 1;
        do {
                if ((err = ftp_xchg(ftp_info, &ind, &last, (char *)data)) < 0 )
                        return err;
                if (ind != 200) {
                        printk(KERN_ERR "FTPFS: *** (---) service error\n");
                        return 9;
                }
        } while (!last);

        if (rest) {
                sprintf(data, "REST %u", rest);
                last = 1;
                do {
                        if ((err = ftp_xchg(ftp_info, &ind, &last, (char *)data)) < 0 )
                                return 0;
                        if (ind != 350) {
                                printk(KERN_ERR "FTPFS: *** (---) service error\n");
                                return 0;
                        }
                } while (!last);
        }

        strcpy(data, command);
        last = 1;
        do {
                if ((err = ftp_xchg(ftp_info, &ind, &last, (char *)data)) < 0 )
                        return err;
                if (ind != 150) {
                        printk(KERN_ERR "FTPFS: *** (---) service error\n");
                        return 10;
                }
        } while (!last);

        if ((err = sock_create(AF_INET, SOCK_STREAM, 0, port_sock)) < 0 ) {
                printk(KERN_ERR "FTPFS: *** (---) create socket error\n");
                return err;
        }

        (*port_sock)->type = ftp_info->port_sock->type;
        
        if ((err = (*port_sock)->ops->dup((*port_sock), ftp_info->port_sock)) < 0) {
                sock_release(*port_sock);
                printk(KERN_ERR "FTPFS: *** (---) dup socket error\n");
                return err;
        }

        if ((err = (*port_sock)->ops->accept(ftp_info->port_sock, (*port_sock), 0)) < 0) {
                sock_release(*port_sock);
                printk(KERN_ERR "FTPFS: *** (---) accept socket error\n");
                return err;
        }

        ftp_info->port_buf_off = 0;
        ftp_info->port_buf_len = 0;

        return 0;
}

int ftp_close(struct ftp_info_type * ftp_info, struct socket * port_sock) {
        int err;
        int ind;
        int last;
        char data[256];

        //while (sock_recv(port_sock, ftp_info->port_buf, ftp_info->port_buf_size) > 0);

        sock_release(port_sock);

        sock_release(ftp_info->port_sock);

        strcpy(data, "ABOR");
        last = 1;
        do {
                if ((err = ftp_xchg(ftp_info, &ind, &last, (char *)data)) < 0 )
                        return err;
                if ((ind != 225) && (ind != 226) && (ind != 425) && (ind != 426) && (ind != 500)) {
                        printk(KERN_ERR "FTPFS: *** (---) service error\n");
                        return 11;
                }
        } while (!last);

        if ((ind == 225) || (ind == 226) || (ind == 425) || (ind == 426)) {
                last = 0;
                do {
                        if ((err = ftp_xchg(ftp_info, &ind, &last, (char *)data)) < 0 )
                                return err;
                        if ((ind != 225) && (ind != 500)) {
                                printk(KERN_ERR "FTPFS: *** (---) service error\n");
                                return 12;
                        }
                } while (!last);
        }

        return 0; 
}

int ftp_recv_port_ascii_data(struct ftp_info_type * ftp_info, struct socket * port_sock, char * data, int * len) {
        int size;
        char temp_char;
        (*len) = 0;
        (*data) = '\000';
        while (1) {
                if (ftp_info->port_buf_len) {
                        temp_char = *((char *)ftp_info->port_buf + ftp_info->port_buf_off);
                        ftp_info->port_buf_off++;
                        ftp_info->port_buf_len--;
                
                        if (temp_char == '\n')
                                return 0;
                        else if (temp_char >= ' ') {
                                (*len)++;
                                if ((*len) < 256) {
                                        (*(data + (*len) - 1)) = temp_char;
                                        (*(data + (*len))) = '\000';
                                }
                                if ((*len) == 256)
                                        printk(KERN_ERR "FTPFS: *** (---) port data too big\n");
                        }
                } else {
                        ftp_info->port_buf_off = 0;
                        if ((size = sock_recv(port_sock, ftp_info->port_buf, ftp_info->port_buf_size)) < 0) {
                                printk(KERN_ERR "FTPFS: *** (---) error receive port data\n");
                                return size;
                        }
                        ftp_info->port_buf_len = size;
                        if (!size)
                                return 0;
                }
        }
}

int dentry_name(struct dentry * dentry, char * name) {
        struct dentry * tmp_dentry = dentry;
        int pos = 0;
        while (tmp_dentry != tmp_dentry->d_parent) {
                pos += tmp_dentry->d_name.len + 1;
                if (pos >= 255 ) {
                        printk(KERN_ERR "FTPFS: *** (---) entry name too long\n");
                        return 13;
                }
                tmp_dentry = tmp_dentry->d_parent;
        };
        *(name + pos) = '\000';
        while (dentry != dentry->d_parent) {
                pos -= dentry->d_name.len;
                strncpy(name + pos, dentry->d_name.name, dentry->d_name.len);
                pos --;
                *(name + pos) = '/';
                dentry = dentry->d_parent;
        };
        if (!(*name)) {
                *name = '/';
                *(name + 1) = '\000';
        }
        return 0;
}

static int ftpfs_readdir(struct file * filp, void * dirent, filldir_t filldir) {
	struct dentry * dentry = filp->f_dentry;
	struct inode * inode = dentry->d_inode;
        struct super_block * s = inode->i_sb;
        struct ftp_info_type * ftp_info = (struct ftp_info_type *)s->u.generic_sbp;
        struct socket * port_sock;
	off_t nr = filp->f_pos;
        off_t cnr;
        char name[256];
        char data[256];
        int len;

        inode->i_atime = CURRENT_TIME;
        
	switch(nr) {
        	case 0:
	        	if (filldir(dirent, ".", 1, nr, inode->i_ino) < 0)
		        	return 0;
                        filp->f_pos = ++nr;
        	case 1:
	        	if (filldir(dirent, "..", 2, nr, inode->i_ino) < 0)
		        	return 0;
                        filp->f_pos = ++nr;
                default:
                        cnr = 2;
                        dentry_name(dentry, (char *)name);
                        sprintf(data, "LIST %s", name);
                        if (!ftp_open(ftp_info, &port_sock, 'A', 0, data))
                                while (1) {
                                        ftp_recv_port_ascii_data(ftp_info, port_sock, (char *)data, &len);
                                        if (!len) {
                                                ftp_close(ftp_info, port_sock);
                                                filp->f_pos = nr;
                                                return 0;
                                        }
                                        if (len > 51)
                                                if ((*(data + 0) == '-') || 
                                                           (
                                                                   (*(data + 0) == 'd') && 
                                                                   (!(*(data + 55) == '.' && *(data + 56) == '\000')) && 
                                                                   (!(*(data + 55) == '.' && *(data + 56) == '.' && *(data + 57) == '\000'))
                                                           )
                                                   ) {
                                                        if (cnr == nr) {
                        	                	        if (filldir(dirent, data + 55, len - 55, nr, inode->i_ino) < 0) {
                                                                        ftp_close(ftp_info, port_sock);
		                                                	return 0;
                                                                }
                                                                nr ++;
                                                        }
                                                        cnr ++;
                                                }
                                }
        }
        return 0;
}

static struct file_operations ftpfs_operations = {
	NULL,                   /* llseek */
        generic_file_read,      /* read */
	NULL,                   /* write */
	ftpfs_readdir,          /* readdir */
	NULL,                   /* poll */
	NULL,			/* ioctl */
	generic_file_mmap,      /* mmap */
	NULL,                   /* open */
	NULL,			/* flush */
	NULL,                   /* release */
	NULL,			/* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL,                	/* revalidate */
	NULL			/* lock */
};

int ftpfs_d_revalidate(struct dentry * dentry, int flags) {
        return (dentry->d_inode != NULL);
}

static struct dentry_operations ftpfs_dentry_operations = {
	ftpfs_d_revalidate,     /* d_revalidate */
	NULL,			/* d_hash */
	NULL,			/* d_compare */
        NULL,                   /* d_delete */
        NULL,                   /* d_release */
        NULL                    /* d_iput */
};

static struct dentry * ftpfs_lookup(struct inode * dir, struct dentry * dentry) {
        struct super_block * s = dir->i_sb;
	dentry->d_op = &ftpfs_dentry_operations;
        dentry->d_inode = iget(s, (unsigned)dentry);
	d_add(dentry, dentry->d_inode);
        return NULL;
}

static int ftpfs_readpage(struct file * file, struct page * page) {
	struct dentry * dentry = file->f_dentry;
	struct inode * inode = dentry->d_inode;
        char name[256];
        int err;
        int ind;
        int last;
        struct super_block * s = inode->i_sb;
        struct ftp_info_type * ftp_info = (struct ftp_info_type *)s->u.generic_sbp;
        struct socket * port_sock;
        char data[256];
        size_t len;
        unsigned long buf;
	int res = -EIO;

        inode->i_atime = CURRENT_TIME;

	atomic_inc(&page->count);
	set_bit(PG_locked, &page->flags);

	buf = page_address(page);
	clear_bit(PG_uptodate, &page->flags);
	clear_bit(PG_error, &page->flags);
	if (page->offset < inode->i_size) {
                len = inode->i_size - page->offset;
                if (len > PAGE_SIZE)
                        len = PAGE_SIZE;
                dentry_name(dentry, (char *)name);
                sprintf(data, "RETR %s", name);
                if (!ftp_open(ftp_info, &port_sock, 'I', page->offset, data)) {
                        size_t size = 0;
                        int retry = 0;
                        while (len - size)
                                if ((res = sock_recv(port_sock, (char *)(buf + size), len - size)) < 0) {
                                        res = -EIO;
                                        break;
                                } else {
                                        if (res) {
                                                retry = 0;
                                                size += res;
                                        
                                        } else {
                                                if (retry ++ < 5)
                                                        printk(KERN_ERR "FTPFS: *** (---) retry %i receive port data\n",retry);
                                                else {
                                                        printk(KERN_ERR "FTPFS: *** (---) error receive port data\n");
                                                        res = -EIO;
                                                        break;                                                
                                                }
                                        }
                                        res = 0;
                                }
                        ftp_close(ftp_info, port_sock);
                }
                if (!res) {
                        if (len < PAGE_SIZE)
				memset((void *)(buf + len),0 , PAGE_SIZE - len);
			set_bit(PG_uptodate, &page->flags);
                }
	}
	if (res) {
		set_bit(PG_error, &page->flags);
		memset((void *)buf, 0, PAGE_SIZE);
	}
	clear_bit(PG_locked, &page->flags);
	wake_up(&page->wait);
	free_page(buf);
	return res;
}

int ftpfs_i_revalidate(struct dentry * dentry) {
	struct inode * inode = dentry->d_inode;
        if (!inode)
                return 0;
        return (inode->i_ctime > inode->i_atime);
}

static struct inode_operations ftpfs_inode_operations = {
	&ftpfs_operations,      /* file operations */
	NULL,                   /* create */
	ftpfs_lookup,           /* lookup */
	NULL,                   /* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,                   /* mknod */
	NULL,                   /* rename */
	NULL,                   /* readlink */
	NULL,                   /* follow_link */
	ftpfs_readpage,         /* readpage */
	NULL,                   /* writepage */
	NULL,                   /* bmap */
	NULL,                   /* truncate */
	NULL,			/* permission */
	NULL,			/* smap */
	NULL,			/* updatepage */
	ftpfs_i_revalidate     	/* revalidate */
};

static void ftpfs_read_inode(struct inode * inode) {
	ino_t ino = inode->i_ino;
        struct super_block * s = inode->i_sb;

        if ((struct super_block *)ino == s) {
        	inode->i_op = &ftpfs_inode_operations;
	        inode->i_mode = S_IFDIR | S_IRUGO | S_IXUGO | S_IWUSR;
        	inode->i_nlink = 2;
	        inode->i_size = 0;
        	inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
	        inode->i_blocks = 0;
        	inode->i_blksize = 1024;
	        inode->i_uid = inode->i_gid = 0;        
        } else {
                struct dentry * dentry;
                char name[256];
                struct ftp_info_type * ftp_info = (struct ftp_info_type *)s->u.generic_sbp;
                struct socket * port_sock;
                char data[256];
                int len;
                if (!inode->u.generic_ip)
                        (struct dentry *)inode->u.generic_ip = (struct dentry *)ino;
                dentry = (struct dentry *)inode->u.generic_ip;
                dentry_name(dentry, (char *)name);

        	inode->i_op = &ftpfs_inode_operations;
                inode->i_mode = 0;
        	inode->i_nlink = 0;
	        inode->i_size = 0;
        	inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
	        inode->i_blocks = 0;
        	inode->i_blksize = 1024;
	        inode->i_uid = inode->i_gid = 0;        

                sprintf(data, "LIST -d %s", name);
                if (!ftp_open(ftp_info, &port_sock, 'A', 0, data)) {
                        int ind;
                        ftp_recv_port_ascii_data(ftp_info, port_sock, (char *)data, &len);
                        ftp_close(ftp_info, port_sock);
                        if (len < 55)
                                printk(KERN_ERR "FTPFS: *** (---) inode read error\n");
                        else {
                                for (ind = 0; ind < 55 ; ind++)
                                        if (*(data + ind) <= ' ')
                                                *(data + ind) = '\000';
                                switch (*(data + 0)) {
                                        case '-':
                                                inode->i_mode |= S_IFREG;
                                                for (ind = 33 ; ind < 41 ; ind ++) {
                                                        inode->i_size *= 10;
                                                        if (*(data + ind))
                                                                inode->i_size += (*(data + ind)) - '0';
                                                }
                                                break;
                                        case 'd':
                                                inode->i_mode |= S_IFDIR;
                                                break;
                                        default:
                                                printk(KERN_ERR "FTPFS: *** (---) unknown file type '%c'\n", *(data + 0));
                                }
                                inode->i_blocks = (inode->i_size + 1023 ) >> 9;
                                if (!strcmp(data + 15, ftp_info->user)) {
                                        if (*(data + 1) == 'r')
                                                inode->i_mode |= S_IRUSR;
                                        if (*(data + 3) == 'x')
                                                inode->i_mode |= S_IXUSR;
                                } else {
                                        if (*(data + 7) == 'r')
                                                inode->i_mode |= S_IRUSR;
                                        if (*(data + 9) == 'x')
                                                inode->i_mode |= S_IXUSR;
                                }
                                for (ind = 11 ; ind < 14 ; ind ++) {
                                        inode->i_nlink *= 10;
                                        if (*(data + ind))
                                                inode->i_nlink += (*(data + ind)) - '0';
                                }
                        }
                }
        }
        return;
}


static void ftpfs_put_super(struct super_block *s) {
        struct ftp_info_type * ftp_info = (struct ftp_info_type *)s->u.generic_sbp;
        char data[256];
        int len;

        ftp_disconnect(ftp_info);

        kfree(ftp_info->port_buf);
        
        kfree(ftp_info->control_buf);
        
        kfree(ftp_info);
        
	MOD_DEC_USE_COUNT;

        return;
}

static int ftpfs_statfs(struct super_block *sb, struct statfs *buf, int bufsiz) {
	struct statfs st;
	memset(&st, 0, sizeof(st));
	st.f_type = FTPFS_MAGIC;
	st.f_bsize = 1024;
	st.f_blocks = 0;
	st.f_namelen = 256;
	return copy_to_user(buf, &st, bufsiz) ? -EFAULT : 0;
}

static struct super_operations ftpfs_sops = {
	ftpfs_read_inode,       /* read_inode */
	NULL,                   /* write_inode */
	NULL,			/* put_inode */
	NULL,			/* delete_inode */
	NULL,			/* notify_change */
	ftpfs_put_super,        /* put_super */
	NULL,			/* write_super */
	ftpfs_statfs,           /* statfs */
	NULL,			/* remount_fs */
	NULL,			/* clear_inode */
};

struct super_block * ftpfs_read_super(struct super_block * s, void * options, int silent) {
        struct ftp_info_type * ftp_info;
        char * temp_ptr;
        char * host_ptr;
        char * port_ptr;
        char * user_ptr;
        char * pass_ptr;
        char * cs_ptr;
        char * ps_ptr;
        int ind;
        int num;
        int last;
        char data[256];


	MOD_INC_USE_COUNT;

	lock_super(s);
	if (s->s_root)
                goto fail1;

        ftp_info = (struct ftp_info_type *)kmalloc(sizeof(*ftp_info), GFP_KERNEL);
        if (!ftp_info)
                goto fail1;

        host_ptr = NULL;
        port_ptr = NULL;
        user_ptr = NULL;
        pass_ptr = NULL;
        cs_ptr = NULL;
        ps_ptr = NULL;
        temp_ptr = NULL;
        if (options)
                temp_ptr = strtok(options,",");
        for ( ; temp_ptr ; temp_ptr = strtok(NULL,",")) {
                if (!strncmp(temp_ptr,"host=",5))
                        host_ptr = temp_ptr+5;
                else if (!strncmp(temp_ptr,"port=",5))
                        port_ptr = temp_ptr+5;
                else if (!strncmp(temp_ptr,"user=",5))
                        user_ptr = temp_ptr+5;
                else if (!strncmp(temp_ptr,"pass=",5))
                        pass_ptr = temp_ptr+5;
                else if (!strncmp(temp_ptr,"cs=",3))
                        cs_ptr = temp_ptr+3;
                else if (!strncmp(temp_ptr,"ps=",3))
                        ps_ptr = temp_ptr+3;
                else
                        printk(KERN_WARNING "FTPFS: --- (---) unknown option (%s)\n",temp_ptr);
        }
        if (host_ptr) {
                ftp_info->control_address.sin_addr.s_addr = 0x00000000;
                temp_ptr = strtok(host_ptr,".");
                for (ind = 0; ind < 32 ; ind+=8) {
                        if (!*temp_ptr) {
                                printk(KERN_ERR "FTPFS: *** (---) invalid host address\n");
                                goto fail2;
                        }
                        num = simple_strtoul(temp_ptr,&temp_ptr,0);
                        if(*temp_ptr) {
                                printk(KERN_ERR "FTPFS: *** (---) invalid host address\n");
                                goto fail2;
                        }
                        if (num < 0 || num > 255) {
                                printk(KERN_ERR "FTPFS: *** (---) invalid host address\n");
                                goto fail2;
                        }
                        ftp_info->control_address.sin_addr.s_addr |= (num << ind);
                        temp_ptr = strtok(NULL,".");
                }
        } else
                ftp_info->control_address.sin_addr.s_addr = 0x0100007f;

        if (port_ptr) {
                num = simple_strtoul(port_ptr,&port_ptr,0);
                if(*port_ptr) {
                        printk(KERN_ERR "FTPFS: *** (---) invalid port\n");
                        goto fail2;
                }
                if (num < 0 || num > 65535) {
                        printk(KERN_ERR "FTPFS: *** (---) invalid port\n");
                        goto fail2;
                }
                ftp_info->control_address.sin_port = ntohs(num);
        } else
                ftp_info->control_address.sin_port = ntohs(21);

        if (user_ptr) {
                if (strlen(user_ptr)>64) {
                        printk(KERN_ERR "FTPFS: *** (---) invalid user name (max 64 chars)\n");
                        goto fail2;
                }
                strcpy(ftp_info->user,user_ptr);
        } else
                strcpy(ftp_info->user,"anonymous");

        if (pass_ptr) {
                if (strlen(pass_ptr)>64) {
                        printk(KERN_ERR "FTPFS: *** (---) invalid user password (max 64 chars)\n");
                        goto fail2;
                }
                strcpy(ftp_info->pass,pass_ptr);
        } else
                strcpy(ftp_info->pass,"nobody");

        if (cs_ptr) {
                num = simple_strtoul(cs_ptr,&cs_ptr,0);
                if(*cs_ptr) {
                        printk(KERN_ERR "FTPFS: *** (---) invalid control buffer size (256..4096)\n");
                        goto fail2;
                }
                if (num < 256 || num > 4096) {
                        printk(KERN_ERR "FTPFS: *** (---) invalid control buffer size (256..4096)\n");
                        goto fail2;
                }
                ftp_info->control_buf_size = num;
        } else
                ftp_info->control_buf_size = 1024;
        ftp_info->control_buf_off = 0;
        ftp_info->control_buf_len = 0;

        if (ps_ptr) {
                num = simple_strtoul(ps_ptr,&ps_ptr,0);
                if(*ps_ptr) {
                        printk(KERN_ERR "FTPFS: *** (---) invalid port buffer size (256..131072)\n");
                        goto fail2;
                }
                if (num < 256 || num > 131072) {
                        printk(KERN_ERR "FTPFS: *** (---) invalid port buffer size (256..131072)\n");
                        goto fail2;
                }
                ftp_info->port_buf_size = num;
        } else
                ftp_info->port_buf_size = 1024;

        ftp_info->control_address.sin_family = AF_INET;

        ftp_info->control_buf = kmalloc(ftp_info->control_buf_size, GFP_KERNEL);
        if (!ftp_info->control_buf)
                goto fail2;

        ftp_info->port_buf = kmalloc(ftp_info->port_buf_size, GFP_KERNEL);
        if (!ftp_info->port_buf)
                goto fail3;

	s->u.generic_sbp = (void *) ftp_info;
	s->s_blocksize = 1024;
	s->s_blocksize_bits = 10;
	s->s_magic = FTPFS_SUPER_MAGIC;
        s->s_flags |= MS_RDONLY;
	s->s_op = &ftpfs_sops;
        s->s_root = d_alloc_root(iget(s, (unsigned)s), NULL);
        if (!s->s_root)
                goto fail4;

        ftp_info->control_sock = NULL;

        if (ftp_connect(ftp_info) != 0 )
                goto fail5;

	unlock_super(s);

        return s;

fail5:
        if (ftp_info->control_sock) {
                sock_release(ftp_info->control_sock);
                ftp_info->control_sock = NULL;
        }

fail4:
        kfree(ftp_info->port_buf);
        
fail3:
        kfree(ftp_info->control_buf);
        
fail2:
        kfree(ftp_info);
     
fail1:
        s->s_dev = (kdev_t)NULL;
        unlock_super(s);
        
        MOD_DEC_USE_COUNT;
        
        printk(KERN_ERR "FTPFS: *** (---) mount failed\n");
        return NULL;
}

static struct file_system_type ftpfs = {
	"ftpfs",                /* name */
	0,                      /* fs_flags */
	ftpfs_read_super,       /* read_super */
	NULL
};

int init_module(void) {
        printk(KERN_INFO "FTPFS: --- (---) 2000, Vladimir E. Spiridenkov <sprite@skin.gtn.ru>\n");
	return register_filesystem(&ftpfs);
}

void cleanup_module( void) {
        unregister_filesystem(&ftpfs);
}
